from .commands import BaseCommand, ClientCommand, ServerCommand, ServerSideClientCommand, ServerSideServerCommand
from .flags import CommandFlag, NetworkFlag

__all__ = [
    "BaseCommand",
    "ClientCommand",
    "ServerCommand",
    "ServerSideClientCommand",
    "ServerSideServerCommand",
    "CommandFlag",
    "NetworkFlag",
]



import socket as _sock
from py_mp.models import ClientBaseModel as _ClientBase
from py_mp.commands import ClientCommand as _ClientCommand, ServerCommand as _ServerCommand, BaseCommand as _BaseCommand, ServerSideClientCommand as _ServerSideClientCommand, ServerSideServerCommand as _ServerSideServerCommand


class NetworkServerBase:
    def __init__(self, host: str | None = None, port: int | None = None, auto_bind: bool = True) -> None:
        """
        Initializes all the variables in the class and prepares them for use.

        Parameters
        ----------
            host: str | None, by default None
                Specify the hostname of the server to bind to
            port: int | None, by default None
                Specify the port to bind to
            auto_bind: bool, by default True
                Automatically bind the socket to a port and host
        """
        self.conn: _sock.socket = _sock.socket(_sock.AF_INET, _sock.SOCK_STREAM)
        self._binded: bool = False
        self.addr: tuple[str, int] | None = None
        self.clients: list[_ClientBase] = []

        # Auto-bind
        if auto_bind:
            if host and port:
                self.bind(host, port)

    def __repr__(self) -> str:
        return f"<NetworkServerBase {f'binded ({self.addr[0]}:{self.addr[1]})' if self.is_binded() else 'not binded'}>"

    def bind(self,  host: str, port: int) -> _sock.socket:
        """Bind the socket to a host and port

        Parameters
        ----------
        host : str
            The Hostname or IP address to bind the server to
        port : int
            The Port to bind the server to

        Returns
        -------
        socket.socket
            The socket object that is binded to the given host and port
        """
        self.addr = (host, port)
        self.conn.bind(self.addr)
        self._binded = True
        return self.conn

    def _accept(self, amount: int = 1) -> None:
        """Wrapper of the socket.accept() method including a check if the socket is binded to a host and port

        Parameters
        ----------
        amount : int, by default 1
            The amount of clients to accept

        Raises
        ------
        ConnectionError
            The socket is not binded to a host and port
        """
        if not self.is_binded():
            raise ConnectionError("Not binded to any addr")
        self.conn.listen(amount)
        while len(self.clients) < amount:
            self.clients.append(_ClientBase.from_accept(*self.conn.accept()))

    def _recv(self, size: int, client: _ClientBase) -> bytes:
        """Wrapper of the socket.recv() method including a check if the socket is binded to a host and port

        Parameters
        ----------
        size : int
            The amount of bytes to receive
        client : ClientBase
            The client to receive the data from

        Returns
        -------
        bytes
            The received bytes

        Raises
        ------
        ConnectionError
            The socket is not binded to a host and port
        """
        if not self.is_binded():
            raise ConnectionError("Not binded to any addr")
        if client not in self.clients:
            raise ConnectionError("Client not connected")
        return client.conn.recv(size)

    def _send(self, data: bytes, client: _ClientBase) -> None:
        """Wrapper of the socket.send() method including a check if the socket is binded to a host and port

        Parameters
        ----------
        data : bytes
            The data to send to the client
        client : ClientBase
            The client to send the data to

        Raises
        ------
        ConnectionError
            The socket is not binded to a host and port
        """
        if not self.is_binded():
            raise ConnectionError("Not binded to any addr")
        if client not in self.clients:
            raise ConnectionError("Client not connected")
        client.conn.send(data)

    def is_binded(self) -> bool:
        """Check if the socket is binded to a host and port

        Returns
        -------
        bool
            True if the socket is binded to a host and port, False if not
        """
        return self._binded


class NetworkServer(NetworkServerBase):
    def __init__(self, *args, **kwargs) -> None:
        """
        Initializes all the variables in the class and prepares them for use.

        The difference between this class and the NetworkServerBase class is that this class
        automatically sends the length of the data to the client and then sends the data
        """
        self.ENCODING: str = "utf-8"
        super().__init__(*args, **kwargs)

    def accept(self, amount: int = 1) -> None:
        """Accept a client

        Parameters
        ----------
        amount : int, by default 1
            The amount of clients to accept
        """
        self._accept(amount)

    def send(self, data: str, client: _ClientBase) -> None:
        """Send data to a specific client

        Parameters
        ----------
        data : bytes
            The data to send to the client
        client : ClientBase
            The client to send the data to
        """
        if client not in self.clients:
            raise ConnectionError("Client not connected")
        data = data.encode(self.ENCODING)
        length = len(data)
        self._send(length.to_bytes(8, "big"), client)
        if int.from_bytes(self._recv(8, client), "big") == length:
            self._send(data, client)

    def send_to(self, data: str, *clients: _ClientBase) -> None:
        """Send data to several clients

        Parameters
        ----------
        data : str
            The data to send to the client
        clients : list[ClientBase]
            The clients to send the data to
        """
        for client in clients:
            self.send(data, client)

    def send_all(self, data: str) -> None:
        """Send data to all clients

        Parameters
        ----------
        data : str
            The data to send to the client
        """
        for client in self.clients:
            self.send(data, client)

    def recv(self, client: _ClientBase) -> str:
        """Receive data from a specific client

        Parameters
        ----------
        client : ClientBase
            The client to receive the data from

        Returns
        -------
        str
            The received data
        """
        if client not in self.clients:
            raise ConnectionError("Client not connected")

        length = int.from_bytes(self._recv(8, client), "big")
        self._send(length.to_bytes(8, "big"), client)
        return self._recv(length, client).decode(self.ENCODING)


class CommandServer(NetworkServer):
    def __init__(self, *args, **kwargs) -> None:
        """
        Initializes all the variables in the class and prepares them for use.

        The difference between this class and the NetworkClient class is that this class
        sends and receives data as strings instead of bytes.
        """
        super().__init__(*args, **kwargs)

    def send(self, command: _ServerCommand | _ServerSideServerCommand, client: _ClientBase) -> None:
        """Send data to the server

        Parameters
        ----------
        command : ClientCommand | ServerCommand
            The command to send to the server
        client : ClientBase
            The client to send the data to
        """
        if isinstance(command, _ServerSideServerCommand):
            command = command.to_client_cmd()
        super().send(command.serialize(), client)

    def send_to(self, command: _ServerCommand | _ServerSideServerCommand, *clients: _ClientBase) -> None:
        """Send data to several clients

        Parameters
        ----------
        command : ServerCommand | ServerSideServerCommand
            The data to send to the client
        clients : list[ClientBase]
            The clients to send the data to
        """
        for client in clients:
            self.send(command, client)

    def send_all(self, command: _ServerCommand | _ServerSideServerCommand) -> None:
        """Send data to all clients

        Parameters
        ----------
        command : ClientCommand | ServerCommand
            The command to send to the server
        """
        if isinstance(command, _ServerSideServerCommand):
            command = command.to_client_cmd()
        super().send_all(command.serialize())

    def recv(self, client: _ClientBase) -> _ServerSideClientCommand:
        """Receive data from the server

        Parameters
        ----------
        client : ClientBase
            The client to receive the data from

        Returns
        -------
        str
            The received data
        """
        return _ServerSideClientCommand.from_client_cmd(_BaseCommand.deserialize(super().recv(client)), client)


if __name__ == '__main__':
    from py_mp.commands import ServerSideServerCommand, NetworkFlag

    server = CommandServer("localhost", 1234)
    server.accept(2)

    print(server.recv(server.clients[0]))
    print(server.recv(server.clients[1]))
    server.send(ServerSideServerCommand(NetworkFlag.CONNECTED, server.clients[0]), server.clients[0])
    server.send(ServerSideServerCommand(NetworkFlag.CONNECTED, server.clients[1]), server.clients[1])



from enum import IntFlag as _IntFlag


class CommandFlag(_IntFlag):
    pass


class NetworkFlag(CommandFlag):
    """
    Flags for the Network Commands
    Reserved Flags: 100 - 199
    """
    DISCONNECTED = 100
    CONNECTED = 101



from dataclasses import dataclass as _dc
import socket as _sock


@_dc
class ClientBaseModel:
    conn: _sock.socket
    address: str
    port: int

    @classmethod
    def from_accept(cls, conn: _sock.socket, addr: tuple[str, int]):
        return cls(conn, addr[0], addr[1])



from .client import NetworkClientBase, NetworkClient, CommandClient
from .server import NetworkServerBase, NetworkServer, CommandServer

__all__ = [
    "NetworkClientBase",
    "NetworkClient",
    "CommandClient",
    "NetworkServerBase",
    "NetworkServer",
    "CommandServer",
]



import json

from py_mp.models import ClientBaseModel as _ClientBase
from py_mp.commands.flags import CommandFlag as _CommandFlag, NetworkFlag as _NetworkFlag


class BaseCommand:
    def __init__(self, flag: _CommandFlag, **kwargs):
        """
        Initializes all the variables in the class and prepares them for use.

        Base Class of any command that is sent/received from/to the server or client.

        Parameters
        ----------
        flag: int
            The flag of the command
        kwargs
            The arguments of the command
        """
        self.flag = flag
        self.args = kwargs

    def __repr__(self):
        return f"<BaseCommand [{self.flag.name}] args: {', '.join(self.args.keys())})>"

    def serialize(self) -> str:
        """
        Serializes the command into a string

        Returns
        -------
        str
            The serialized command
        """
        return json.dumps({"flag": self.flag.value, "args": self.args})

    @classmethod
    def deserialize(cls, serial: str) -> "BaseCommand":
        """
        Deserializes a string into a command

        Parameters
        ----------
        serial: str
            The serialized command

        Returns
        -------
        BaseCommand
            The deserialized command
        """
        data = json.loads(serial)
        if _NetworkFlag(data.get("flag")) in _NetworkFlag:
            flg = _NetworkFlag(data.get("flag"))
        else:
            flg = _CommandFlag(data.get("flag"))
        return cls(flg, **data.get("args", {}))


class ClientCommand(BaseCommand):
    """
    Client-Side (Client) Command

    Instance that reprs a command that is sent from the client to the server
    """
    def __repr__(self):
        return f"<ClientCommand [{self.flag.name}] args: {', '.join(self.args.keys())})>"


class ServerCommand(BaseCommand):
    """
    Client-Side (Server) Command

    Instance that reprs a command that is sent from the server to the client
    """
    def __repr__(self):
        return f"<ServerCommand [{self.flag.name}] args: {', '.join(self.args.keys())})>"


class ServerSideClientCommand(BaseCommand):
    def __init__(self, flag: _CommandFlag, client: _ClientBase, **kwargs):
        """
        Initializes all the variables in the class and prepares them for use.

        Server-Side (Client) Command
        Instance that reprs a command that is sent from the client to a server

        Parameters
        ----------
        flag : CommandFlag
            The flag for the command
        client : ClientBase
            The client that sent the command
        kwargs
            The arguments sent from the client
        """
        super().__init__(flag, **kwargs)
        self.client: _ClientBase = client

    def __repr__(self):
        return f"<ServerSideClientCommand [{self.flag.name}] args: {', '.join(self.args.keys())})>"

    @classmethod
    def from_client_cmd(cls, cmd: BaseCommand, client: _ClientBase) -> "ServerSideClientCommand":
        """
        Creates a ServerSideClientCommand from a ClientCommand

        Parameters
        ----------
        cmd: BaseCommand
            The Command the client sent
        client: ClientBase
            The client that sent the command

        Returns
        -------
        ServerSideClientCommand
            The Command including the object of the client that sent it
        """
        return cls(cmd.flag, client, **cmd.args)


class ServerSideServerCommand(BaseCommand):
    def __init__(self, flag: _CommandFlag, client: _ClientBase, **kwargs):
        """
        Initializes all the variables in the class and prepares them for use.

        Server-Side (Server) Command
        Instance that reprs a command that is sent from the server to a client

        Parameters
        ----------
        flag: CommandFlag
            The flag of the command
        client: ClientBase
            The client that the command is being sent to
        kwargs
            Additional arguments to be sent with the command
        """
        super().__init__(flag, **kwargs)
        self.client: _ClientBase = client

    def __repr__(self):
        return f"<ServerSideServerCommand [{self.flag.name}] args: {', '.join(self.args.keys())})>"

    def to_client_cmd(self) -> ServerCommand:
        """
        Creates a ServerCommand from a ServerSideServerCommand

        Returns
        -------
        ServerCommand
            The Command without the object of the client that will be sent to the client
        """
        return ServerCommand(self.flag, **self.args)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['py_mp', 'py_mp.commands', 'py_mp.models', 'py_mp.network']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'python-multiplayer',
    'version': '0.1.2',
    'description': 'Server Client Structure for Python',
    'long_description': '# Python Multiplayer\n\n<a href="https://pypi.org/project/python-multiplayer/"><img src="https://badge.fury.io/py/python-multiplayer.png" alt="PyPI version" height="18"></a>\n\nFramework for Client Server Structure in Python.\nIs intended to be used for multiplayer games in pygame with this Module.\n\n[pygame-multiplayer](https://github.com/BroCodeAT/python-multiplayer)\n\n----\n\n## Planned features\n- [x] Command based Network\n- [ ] Threaded Network\n- [ ] Async Network\n\n\n## Installation\n\n```bash\npip install python-multiplayer\n```\n\n## Usage (Command Server/Client)\n\nServer:\n```python\nfrom py_mp import CommandServer\nfrom py_mp import ServerSideServerCommand\nfrom py_mp.commands import NetworkFlag\n\nserver = CommandServer("localhost", 5000)\nserver.accept()\n\n# Receive a Command from the Client\ncom = server.recv(server.clients[0])\nprint(com)\n\n# Send a Test Command back to the Client\nserver.send(\n    ServerSideServerCommand(NetworkFlag.CONNECTED, server.clients[0], test="test"), \n    server.clients[0]\n)\n```\nClient:\n```python\nfrom py_mp import CommandClient\nfrom py_mp import ClientCommand\nfrom py_mp.commands import NetworkFlag\n\nclient = CommandClient("localhost", 5000)\n\n# Send a Test Command to the Server\nclient.send(ClientCommand(NetworkFlag.CONNECTED, test="test"))\n\n# Receive a Command from the Server\ncom = client.recv()\nprint(com)\n```\n',
    'author': 'dpfurners',
    'author_email': 'dpfurner@tsn.at',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/BroCodeAT/python-multiplayer',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from .network import NetworkServer, NetworkClient, CommandClient, CommandServer
from .commands import ClientCommand, ServerCommand, ServerSideClientCommand, ServerSideServerCommand

__version__ = "0.1.2"



import socket as _sock
from py_mp.commands import ClientCommand as _ClientCommand, \
    ServerCommand as _ServerCommand, BaseCommand as _BaseCommand


class NetworkClientBase:
    def __init__(self, host: str | None = None, port: int | None = None, auto_connect: bool = True) -> None:
        """
        Initializes all the variables in the class and prepares them for use.

        Parameters
        ----------
            host: str | None, by default None
                Specify the hostname of the server to connect to
            port: int | None, by default None
                Specify the port to connect to
            auto_connect: bool, by default True
                Automatically connect the socket to a port and host
        """
        self.conn: _sock.socket = _sock.socket(_sock.AF_INET, _sock.SOCK_STREAM)
        self._connected: bool = False
        self.addr: tuple[str, int] | None = None

        if auto_connect:
            if host and port:
                self.connect(host, port)

    def __repr__(self) -> str:
        return f"<NetworkClientBase " \
               f"{f'connected ({self.addr[0]}:{self.addr[1]})' if self.is_connected() else 'not connected'}>"

    def connect(self,  host: str, port: int):
        """Connect the socket to a host and port

        Parameters
        ----------
        host : str
            The Hostname or IP address of the server to connect to
        port : int
            The Port of the server to connect to

        Returns
        -------
        socket.socket
            The socket object that is connected to the given host and port
        """
        self.addr = (host, port)
        self.conn.connect(self.addr)
        self._connected = True

    def _recv(self, size: int) -> bytes:
        """Wrapper of the socket.recv() method including a check if the socket is connected to a host and port

        Parameters
        ----------
        size : int
            The amount of bytes to receive

        Returns
        -------
        bytes
            The received bytes

        Raises
        ------
        ConnectionError
            The socket is not connected to a host and port
        """
        if self.is_connected():
            return self.conn.recv(size)
        else:
            raise ConnectionError("Not connected to any server")

    def _send(self, data: bytes):
        """Wrapper of the socket.send() method including a check if the socket is connected to a host and port

        Parameters
        ----------
        data : bytes
            The data to send to the server

        Raises
        ------
        ConnectionError
            The socket is not connected to a host and port
        """
        if self.is_connected():
            self.conn.send(data)
        else:
            raise ConnectionError("Not connected to any server")

    def is_connected(self) -> bool:
        """Check if the socket is connected to a host and port (server)

        Returns
        -------
        bool
            True if the socket is connected to a host and port, False if not
        """
        return self._connected


class NetworkClient(NetworkClientBase):
    def __init__(self, *args, **kwargs) -> None:
        """
        Initializes all the variables in the class and prepares them for use.

        The difference between this class and the NetworkClientBase class is that this class
        first sends the size of the data to the server before sending the data itself.
        """
        self.ENCODING: str = "utf-8"
        super().__init__(*args, **kwargs)

    def send(self, data: str):
        """Send data to the server

        Parameters
        ----------
        data : bytes
            The data to send to the server
        """
        data = data.encode(self.ENCODING)
        length = len(data)
        self._send(length.to_bytes(8, "big"))
        if int.from_bytes(self._recv(8), "big") == length:
            self._send(data)

    def recv(self) -> str:
        """Receive data from the server

        Returns
        -------
        str
            The received data
        """
        length = int.from_bytes(self._recv(8), "big")
        self._send(length.to_bytes(8, "big"))
        return self._recv(length).decode(self.ENCODING)


class CommandClient(NetworkClient):
    def __init__(self, *args, **kwargs) -> None:
        """
        Initializes all the variables in the class and prepares them for use.

        The difference between this class and the NetworkClient class is that this class
        sends and receives data as strings instead of bytes.
        """
        super().__init__(*args, **kwargs)

    def send(self, command: _ClientCommand | _ServerCommand):
        """Send data to the server

        Parameters
        ----------
        command : ClientCommand | ServerCommand
            The command to send to the server
        """
        super().send(command.serialize())

    def recv(self) -> _BaseCommand:
        """Receive data from the server

        Returns
        -------
        str
            The received data
        """
        return _BaseCommand.deserialize(super().recv())


if __name__ == '__main__':
    from py_mp.commands import NetworkFlag

    client = CommandClient("localhost", 1234)
    client.send(_ClientCommand(NetworkFlag.CONNECTED, test="test"))
    print(client.recv())
